import pytest
import httpx
from unittest.mock import ANY
from fastapi.testclient import TestClient
from authlib.integrations.base_client.errors import MismatchingStateError
from requests import Response

from geomagio.api.secure import app
from geomagio.api.secure.SessionMiddleware import SessionMiddleware
from geomagio.api.secure.login import User

client = TestClient(app)


@pytest.fixture()
def valid_token():
    return {
        "access_token": "test_access_token",
        "token_type": "Bearer",
        "expires_in": 7200,
        "refresh_token": "test_refresh_toekn",
        "scope": "openid email profile",
        "created_at": 1733244369,
        "id_token": "test_id_token",
        "expires_at": 1733251569,
        "userinfo": {
            "iss": "http://test_url",
            "sub": "1234",
            "aud": "test_aud",
            "exp": 1733244489,
            "iat": 1733244369,
            "nonce": "test_nonce",
            "auth_time": 1733244368,
            "sub_legacy": "test_sub_legacy",
            "name": "test_user",
            "nickname": "test_user_nickname",
            "preferred_username": "test_preferred_username",
            "email": "test_email",
            "email_verified": True,
            "profile": "http://test_url/user",
            "picture": "http://picture_url",
            "groups_direct": ["group1", "group2"],
        },
    }


@pytest.fixture()
def valid_userinfo():
    return {
        "sub": "1234",
        "sub_legacy": "test_sub_legacy",
        "name": "test_user",
        "nickname": "test_user_nickname",
        "preferred_username": "test_preferred_username",
        "email": "test_email",
        "email_verified": True,
        "profile": "http://test_url/user",
        "picture": "http://picture_url",
        "groups": ["group1", "group2"],
    }


@pytest.fixture()
def valid_gitlab_user():
    return {
        "id": 1234,
        "email": "test_email",
        "name": "test_user",
        "username": "test_user_nickname",
        "avatar_url": "http://picture_url",
    }


@pytest.mark.asyncio
async def test_authorization_valid(valid_token, valid_userinfo, mocker):

    mock_authorize = mocker.AsyncMock(return_value=valid_token)
    mocker.patch(
        "geomagio.api.secure.login.oauth.openid.authorize_access_token", mock_authorize
    )

    mock_userinfo = mocker.AsyncMock(return_value=valid_userinfo)
    mocker.patch("geomagio.api.secure.login.oauth.openid.userinfo", mock_userinfo)

    mock_save_session = mocker.AsyncMock()
    mocker.patch.object(SessionMiddleware, "save_session", mock_save_session)

    response = client.get("/authorize")
    assert response.status_code == 200

    valid_session = {
        "token": valid_token,
        "user": valid_userinfo,
    }

    # assert the session data is correct
    mock_save_session.assert_called_once_with(ANY, valid_session)


@pytest.mark.asyncio
async def test_authorization_invalid_authorize_access_token():
    # assert that requests directly to GET /authorize cause an exception
    with pytest.raises(MismatchingStateError) as err:
        client.get("/authorize")

    assert (
        str(err.value)
        == "mismatching_state: CSRF Warning! State not equal in request and response."
    )


@pytest.mark.asyncio
async def test_login_redirects(mocker):
    mock_save_session = mocker.AsyncMock()
    mocker.patch.object(SessionMiddleware, "save_session", mock_save_session)

    mock_redirect = mocker.AsyncMock(return_value={"status_code": 302})
    mocker.patch(
        "geomagio.api.secure.login.oauth.openid.authorize_redirect", mock_redirect
    )

    header = "https://testserver/ws/secure/metadata/1"
    client.get("/login", headers={"Referer": header})

    # assert oauth is called with the correct redirect url
    mock_redirect.assert_called_once_with(
        ANY, "https://testserver/ws/secure/authorize", prompt="login"
    )

    # assert session is saved with correct referrer header
    mock_save_session.assert_called_once_with(ANY, {"after_authorize_redirect": header})


@pytest.mark.asyncio
async def test_logout(valid_token, valid_userinfo, mocker):
    valid_session = {"token": valid_token, "user": valid_userinfo}
    mock_get_session = mocker.AsyncMock(return_value=valid_session)
    mocker.patch.object(SessionMiddleware, "get_session", mock_get_session)

    mock_delete_session = mocker.AsyncMock()
    mocker.patch.object(SessionMiddleware, "delete_session", mock_delete_session)

    client.get(url="/logout", headers={"Cookie": "PHPSESSID=valid_session_id"})

    mock_get_session.assert_called_once_with("valid_session_id")
    mock_delete_session.assert_called_once_with("valid_session_id")


@pytest.mark.asyncio
async def test_user_with_valid_session(valid_token, valid_userinfo, mocker):
    valid_session = {"token": valid_token, "user": valid_userinfo}
    mock_get_session = mocker.AsyncMock(return_value=valid_session)
    mocker.patch.object(SessionMiddleware, "get_session", mock_get_session)

    mock_save_session = mocker.AsyncMock(return_value=valid_session)
    mocker.patch.object(SessionMiddleware, "save_session", mock_save_session)

    response = client.get(url="/user", headers={"Cookie": "PHPSESSID=valid_session_id"})

    assert response.status_code == 200
    mock_get_session.assert_called_once_with("valid_session_id")

    user = User(**response.json())
    assert user.email == "test_email"
    assert user.sub == 1234
    assert user.groups == ["group1", "group2"]


@pytest.mark.asyncio
async def test_user_with_no_session_and_valid_token(
    valid_token, valid_userinfo, valid_gitlab_user, mocker
):
    valid_user_response = httpx.Response(status_code=200, json=valid_gitlab_user)

    valid_groups = [
        {
            "id": 1,
            "full_path": "group1",
        },
        {"id": 2, "full_path": "group2"},
    ]

    valid_groups_response = httpx.Response(status_code=200, json=valid_groups)

    mock_gitlab_request = mocker.AsyncMock(
        side_effect=[
            valid_user_response,
            valid_groups_response,
            httpx.Response(status_code=200, json={}),
        ]
    )
    mocker.patch("httpx.AsyncClient.get", mock_gitlab_request)

    valid_session = {"token": valid_token, "user": valid_userinfo}

    mock_save_session = mocker.AsyncMock(return_value=valid_session)
    mocker.patch.object(SessionMiddleware, "save_session", mock_save_session)

    response = client.get(url="/user", headers={"Authorization": "valid_gitlab_token"})

    session_user = {
        "email": "test_email",
        "sub": 1234,
        "groups": ["group1", "group2"],
        "name": "test_user",
        "nickname": "test_user_nickname",
        "picture": "http://picture_url",
    }

    assert response.status_code == 200
    mock_save_session.assert_called_once_with(ANY, {"user": session_user})

    user = User(**response.json())
    assert user.email == "test_email"
    assert user.sub == 1234
    assert user.groups == ["group1", "group2"]


@pytest.mark.asyncio
async def test_user_with_no_session(mocker):
    mock_get_session = mocker.AsyncMock()
    mocker.patch.object(SessionMiddleware, "get_session", mock_get_session)

    # do not pass in cookie
    response = client.get(url="/user")
    assert response.status_code == 401

    mock_get_session.assert_not_called()


@pytest.mark.asyncio
async def test_user_with_no_session_and_invalid_token(mocker):
    mock_get_session = mocker.AsyncMock()
    mocker.patch.object(SessionMiddleware, "get_session", mock_get_session)

    mock_gitlab_request = mocker.AsyncMock(return_value=None)
    mocker.patch("geomagio.api.secure.login.get_gitlab_user", mock_gitlab_request)

    # do not pass in cookie
    response = client.get(
        url="/user", headers={"Authorization": "invalid_gitlab_token"}
    )
    assert response.status_code == 401

    mock_get_session.assert_not_called()
    mock_gitlab_request.assert_called_once_with(token="invalid_gitlab_token")


@pytest.mark.asyncio
async def test_user_invalid_session(mocker):
    # mock invalid session. this is created when users GET /metadata without logging in
    invalid_session = {
        "redirect_uri": "redirect_uri",
        "nonce": "nonce_str",
        "url": "test_url",
    }
    mock_get_session = mocker.AsyncMock(return_value=invalid_session)
    mocker.patch.object(SessionMiddleware, "get_session", mock_get_session)

    mock_save_session = mocker.AsyncMock()
    mocker.patch.object(SessionMiddleware, "save_session", mock_save_session)

    response = client.get(
        url="/user", headers={"Cookie": "PHPSESSID=invalid_session_id"}
    )

    assert response.status_code == 401
    mock_get_session.assert_called_once_with("invalid_session_id")
